@@ -14,7 +14,7 @@ import logging
 import sys
 import traceback
 
-from raven._compat import string_types, iteritems, text_type
+from raven.utils.compat import string_types, iteritems, text_type
 from raven.base import Client
 from raven.utils.encoding import to_string
 from raven.utils.stacks import iter_stack_frames
@@ -26,6 +26,37 @@ RESERVED = frozenset((
 ))
 
 
+CONTEXTUAL = frozenset((
+    'user', 'culprit', 'server_name', 'fingerprint'
+))
+
+
+def extract_extra(record, reserved=RESERVED, contextual=CONTEXTUAL):
+    data = {}
+
+    extra = getattr(record, 'data', None)
+    if not isinstance(extra, dict):
+        if extra:
+            extra = {'data': extra}
+        else:
+            extra = {}
+    else:
+        # record.data may be something we don't want to mutate to not cause unexpected side effects
+        extra = dict(extra)
+
+    for k, v in iteritems(vars(record)):
+        if k in reserved:
+            continue
+        if k.startswith('_'):
+            continue
+        if '.' not in k and k not in contextual:
+            extra[k] = v
+        else:
+            data[k] = v
+
+    return data, extra
+
+
 class SentryHandler(logging.Handler, object):
     def __init__(self, *args, **kwargs):
         client = kwargs.get('client_cls', Client)
@@ -104,24 +135,7 @@ class SentryHandler(logging.Handler, object):
         return frames
 
     def _emit(self, record, **kwargs):
-        data = {}
-
-        extra = getattr(record, 'data', None)
-        if not isinstance(extra, dict):
-            if extra:
-                extra = {'data': extra}
-            else:
-                extra = {}
-
-        for k, v in iteritems(vars(record)):
-            if k in RESERVED:
-                continue
-            if k.startswith('_'):
-                continue
-            if '.' not in k and k not in ('culprit', 'server_name', 'fingerprint'):
-                extra[k] = v
-            else:
-                data[k] = v
+        data, extra = extract_extra(record)
 
         stack = getattr(record, 'stack', None)
         if stack is True:
@@ -162,13 +176,15 @@ class SentryHandler(logging.Handler, object):
         data['level'] = record.levelno
         data['logger'] = record.name
 
-        if hasattr(record, 'tags'):
-            kwargs['tags'] = record.tags
-        elif self.tags:
-            kwargs['tags'] = self.tags
+        kwargs['tags'] = tags = {}
+        if self.tags:
+            tags.update(self.tags)
+        tags.update(getattr(record, 'tags', {}))
 
         kwargs.update(handler_kwargs)
+        sample_rate = extra.pop('sample_rate', None)
 
         return self.client.capture(
             event_type, stack=stack, data=data,
-            extra=extra, date=date, **kwargs)
+            extra=extra, date=date, sample_rate=sample_rate,
+            **kwargs)
